create table frc_employees
   as
      select *
        from employees
/

create or replace package frc_demo 
is
   type last_names_aat is
      table of frc_employees.last_name%type index by pls_integer;
      
   type employees_aat is
      table of frc_employees%ROWTYPE index by pls_integer;
      
   function last_name (
      employee_id_in in frc_employees.employee_id%type
   ) return frc_employees.last_name%type
      result_cache;

   function last_names (
      department_id_in in frc_employees.department_id%type
   ) return last_names_aat
      result_cache;

   function employees_in_dept (
      department_id_in in frc_employees.department_id%type
   ) return employees_aat
      result_cache;
end;
/

create or replace package body frc_demo is
   function last_name (
      employee_id_in in frc_employees.employee_id%type
   ) return frc_employees.last_name%type
      result_cache
   is
      l_name frc_employees.last_name%type;
   begin
      dbms_output.put_line('EXECUTED BODY OF FUNCTION: got last name for ID ' || employee_id_in);
      select last_name
        into l_name
        from frc_employees
       where employee_id = employee_id_in;

      return l_name;
   end;

   function last_names (
      department_id_in in frc_employees.department_id%type
   ) return last_names_aat
      result_cache
   is
      l_names last_names_aat;
   begin
      dbms_output.put_line('Getting all last names for department ' || department_id_in);
      select last_name
        bulk collect
        into l_names
        from frc_employees
       where department_id = department_id_in;

      return l_names;
   end;

   function employees_in_dept (
      department_id_in in frc_employees.department_id%type
   ) return employees_aat
      result_cache
   is
      l_employees employees_aat;
   begin
      dbms_output.put_line('Getting all rows for department ' || department_id_in);
      
      select *
        bulk collect
        into l_employees
        from frc_employees
       where department_id = department_id_in;

      return l_employees;
   end;
end;
/

/*
Demonstration of scalar caching
*/

begin
   for indx in 1..10 loop
      dbms_output.put_line(frc_demo.last_name(138));
   end loop;
end;
/

/*
Can I cache an entire collection of strings? Yes!
*/

declare
   l_names frc_demo.last_names_aat;
begin
   l_names := frc_demo.last_names(50);
   dbms_output.put_line('1. Count of names in department 50 = ' || l_names.COUNT);
   /*
   Querying all rows into a collection does NOT help with caching
   based on a single employee ID, even if the data returned is the "same"
   and has been cached by another function.
   */
   dbms_output.put_line('Last name for 198 = ' || frc_demo.last_name(198));
   l_names := frc_demo.last_names(50);
   l_names := frc_demo.last_names(60);
   dbms_output.put_line('2. Count of names in department 60 = ' || l_names.COUNT);
end;
/

/*
Can I cache an entire collection of records? Yes!
*/

declare
   l_employees frc_demo.employees_aat;
begin
   l_employees := frc_demo.employees_in_dept(50);
   dbms_output.put_line('1. Count of names in department 50 = ' || l_employees.COUNT);
   l_employees := frc_demo.employees_in_dept(50);
   l_employees := frc_demo.employees_in_dept(60);
   dbms_output.put_line('2. Count of names in department 60 = ' || l_employees.COUNT);
end;
/

/*
Will the cache show "dirty" data in the same session? NO!
*/

set serveroutput on

begin
   dbms_output.put_line('BEFORE update 1: ' || frc_demo.last_name(144));
   dbms_output.put_line('BEFORE update 2: ' || frc_demo.last_name(144));
   dbms_output.put_line('-');
   update frc_employees
      set last_name = 'Uh-oh'
    where employee_id = 198;

   /* Shows me the new value and caching is disabled. */
   dbms_output.put_line('AFTER update 1: ' || frc_demo.last_name(144));
   dbms_output.put_line('AFTER update 2: ' || frc_demo.last_name(144));
   dbms_output.put_line('-');

   /* After commit, will it start caching again? Yes! */
   commit;
   dbms_output.put_line('AFTER commit 1: ' || frc_demo.last_name(144));
   dbms_output.put_line('AFTER commit 2: ' || frc_demo.last_name(144));
   dbms_output.put_line('-');

   /* Reset last name */
   update frc_employees
      set last_name = 'OConnell'
    where employee_id = 198;

   commit;
end;
/

/*
Demonstration that the cache is NOT necessarily
related to SQL operations in the function....
*/

create or replace package body emplu11g is
   function onerow (
      employee_id_in in employees.employee_id%type
   ) return employees%ROWTYPE
      result_cache
   is
      onerow_rec employees%ROWTYPE;
   begin
      onerow_rec.last_name   := get_name(employee_id_in);
      onerow_rec.first_name  := 'Handsome';
      onerow_rec.employee_id := employee_id_in;
      return onerow_rec;
   end;
end;
/

/* Use for app config / global values optimization */

create table app_config (
   app_state varchar2(100)
)
/

begin
   insert into app_config values ( 'DEVELOPMENT' );
   commit;
end;
/

create or replace function in_production return boolean is
   l_state app_config.app_state%type;
begin
   select app_state
     into l_state
     from app_config;
   return l_state = 'PRODUCTION';
end;
/

declare
   l_count   integer := 0;
   l_in_prod boolean;
begin
   tmr.start_timer;
   for indx in 1..10000 loop -- 22 secs for 1000000
      l_in_prod := in_production();
   end loop;
   tmr.show_elapsed('No FRC');
end;
/

create or replace function in_production return boolean result_cache is
   l_state app_config.app_state%type;
begin
   select app_state
     into l_state
     from app_config;
   return l_state = 'PRODUCTION';
end;
/

declare
   l_count   integer := 0;
   l_in_prod boolean;
begin
   tmr.start_timer;
   for indx in 1..10000 loop -- .3 secs for 1000000
      l_in_prod := in_production();
   end loop;
   tmr.show_elapsed('With FRC');
end;
/