/*
Aggregates and their necessary adjunct, qualified expressions, improve program clarity and programmer productivity.

Through Oracle Database 12c release 2, it was possible to supply the value of a non-scalar datatype with an expression, 
for example by using the type constructor for a nested table. Starting with Oracle Database Release 18c, any PL/SQL 
value can be provided by an expression (for example for a record or for an associative array) like a constructor 
provides an abstract datatype value. In PL/SQL, we use the terms "qualified expression" and "aggregate" rather 
than the SQL term "type constructor", but the functionality is the same.

Qualified expressions improve program clarity and developer productivity by providing the ability to declare and 
define a complex value in a compact form where the value is needed.

A qualified expression combines expression elements to create values of a RECORD type or associative array type.

Qualified expressions use an explicit type indication to provide the type of the qualified item. This explicit 
indication is known as a typemark.
*/
DECLARE
   TYPE species_rt IS RECORD
   (
      species_name VARCHAR2 (100),
      habitat_type VARCHAR2 (100),
      surviving_population integer
   );
   
   l_elephant   species_rt := species_rt ('Elephant', 'Savannah', '10000');

   PROCEDURE display_species (species_in species_rt default species_rt ('Not Set', 'Global', 0))
   IS
   BEGIN
      dbms_output.put_line ('Species: '||species_in.species_name);
      dbms_output.put_line ('Habitat: '||species_in.habitat_type);
      DBMS_OUTPUT.put_line ('# Left: '||species_in.surviving_population);
   END;
BEGIN
   display_species (l_elephant);
   
   /* Use the default */
   display_species ();
END;
/

/* With named notation */


DECLARE
   TYPE species_rt IS RECORD
   (
      species_name VARCHAR2 (100),
      habitat_type VARCHAR2 (100),
      surviving_population integer
   );
   
   l_elephant   species_rt := 
      species_rt (species_name => 'Elephant', surviving_population =>'10000', habitat_type => 'Savannah');

   PROCEDURE display_species (species_in species_rt default species_rt ('Not Set', 'Global', 0))
   IS
   BEGIN
      dbms_output.put_line ('Species: '||species_in.species_name);
      dbms_output.put_line ('Habitat: '||species_in.habitat_type);
      DBMS_OUTPUT.put_line ('# Left: '||species_in.surviving_population);
   END;
BEGIN
   display_species (l_elephant);
   
   /* Use the default */
   display_species ();
END;
/

/*
Species: Elephant
Habitat: Savannah
# Left: 10000
Species: Not Set
Habitat: Global
# Left: 0
*/

/* Constructor function with simple list of scalars 

ORA-06550: line 5, column 31:
PLS-00681: named association syntax is required 

*/

DECLARE
   TYPE ints_t IS TABLE OF INTEGER
      INDEX BY PLS_INTEGER;

   l_ints   ints_t := ints_t (55, 555, 5555);
BEGIN
   NULL;
END;
/

/* MUST use named notation! */

DECLARE
   TYPE ints_t IS TABLE OF INTEGER
      INDEX BY PLS_INTEGER;

   l_ints   ints_t := ints_t (1 => 55, 2 => 555, 3 => 5555);
BEGIN
   FOR indx IN 1 .. l_ints.COUNT
   LOOP
      DBMS_OUTPUT.put_line (l_ints (indx));
   END LOOP;
END;
/

/* Constructor sparse array */

DECLARE
   TYPE ints_t IS TABLE OF INTEGER
      INDEX BY PLS_INTEGER;

   l_ints   ints_t := ints_t (600 => 55, -5 => 555, 200000 => 5555);
   l_index pls_integer := l_ints.first;
BEGIN
   WHILE l_index IS NOT NULL
   LOOP
      DBMS_OUTPUT.put_line (l_index || ' => ' || l_ints (l_index));
      l_index := l_ints.NEXT (l_index);
   END LOOP;
END;
/

/*
-5 => 555
600 => 55
200000 => 5555
*/

/* Not just for the declaration section */

DECLARE
   TYPE ints_t IS TABLE OF INTEGER
      INDEX BY PLS_INTEGER;

   l_ints   ints_t;
   l_index pls_integer := l_ints.first;
BEGIN
   l_ints := ints_t (600 => 55, -5 => 555, 200000 => 5555);
   WHILE l_index IS NOT NULL
   LOOP
      DBMS_OUTPUT.put_line (l_index || ' => ' || l_ints (l_index));
      l_index := l_ints.NEXT (l_index);
   END LOOP;
END;
/

/* Not just for the declaration section - unless declared a constant */

DECLARE
   TYPE ints_t IS TABLE OF INTEGER
      INDEX BY PLS_INTEGER;

   l_ints constant ints_t;
   l_index pls_integer := l_ints.first;
BEGIN
   NULL;
END;
/

/* Use both record and array constructor functions! */

DECLARE
   TYPE species_rt IS RECORD
   (
      species_name VARCHAR2 (100),
      habitat_type VARCHAR2 (100),
      surviving_population INTEGER
   );
   
   TYPE species_t IS TABLE OF species_rt
      INDEX BY PLS_INTEGER;

   l_species   species_t := 
      species_t (
         2 => species_rt ('Elephant', 'Savannah', '10000'), 
         1 => species_rt ('Dodos', 'Mauritius', '0'), 
         3 => species_rt ('Venus Flytrap', 'North Carolina', '250'));
BEGIN
   FOR indx IN 1 .. l_species.COUNT
   LOOP
      DBMS_OUTPUT.put_line (l_species (indx).species_name);
   END LOOP;
END;
/

/*
Dodos
Elephant
Venus Flytrap
*/

/* Type compatibility: must be SAME TYPE 

ORA-06550: line 16, column 7:
PLS-00382: expression is of wrong type 

*/


DECLARE
   TYPE species_rt IS RECORD
   (
      species_name VARCHAR2 (100),
      habitat_type VARCHAR2 (100),
      surviving_population INTEGER
   );
   
   TYPE species_t1 IS TABLE OF species_rt
      INDEX BY PLS_INTEGER;
   
   TYPE species_t2 IS TABLE OF species_rt
      INDEX BY PLS_INTEGER;

   l_species   species_t1 := 
      species_t2 (
         1 => species_rt ('Elephant', 'Savannah', '10000'));
BEGIN
   NULL;
END;
/