CREATE OR REPLACE PACKAGE date_mgr
IS
   c_date_mask             CONSTANT VARCHAR2 (100) := 'YYYY-MM-DD HH24:MI:SS';
   c_minutes_in_day        CONSTANT PLS_INTEGER := 24 * 60;
   c_seconds_in_day        CONSTANT PLS_INTEGER := c_minutes_in_day * 60;
   
   FUNCTION iw_to_date (iw IN VARCHAR2)
      RETURN DATE;

   FUNCTION q_to_date (q IN VARCHAR2)
      RETURN DATE;

   FUNCTION first_day_in_month (date_in IN DATE)
      RETURN DATE;

   FUNCTION start_of_day (date_in IN DATE DEFAULT SYSDATE)
      RETURN DATE;

   FUNCTION end_of_day (date_in IN DATE DEFAULT SYSDATE)
      RETURN DATE;

   FUNCTION start_of_week (date_in IN DATE DEFAULT SYSDATE)
      RETURN DATE;

   FUNCTION end_of_week (date_in IN DATE DEFAULT SYSDATE)
      RETURN DATE;

   FUNCTION previous_quarter (quarter_in IN VARCHAR2)
      RETURN VARCHAR2;

   FUNCTION previous_month (month_in IN VARCHAR2)
      RETURN VARCHAR2;

   /* Convert day of month to Nth format */
   FUNCTION dd_to_nth (dd_in IN NUMBER)
      RETURN VARCHAR2;

   /* Convert integer to Nth format */
   FUNCTION n_to_nth (n_in IN NUMBER)
      RETURN VARCHAR2;

   FUNCTION nth_from_n (n_in IN NUMBER)
      RETURN VARCHAR2;

   FUNCTION date_at_time (date_in IN DATE)
      RETURN VARCHAR2;

   /* Offsets can be N = Now, T = Tomorrow, W = Start of Next week */
   FUNCTION date_from_offset (
      offset_in   IN VARCHAR2,
      date_in     IN DATE DEFAULT NULL)
      RETURN DATE;

   FUNCTION q_as_number (date_in IN DATE)
      RETURN NUMBER;

   FUNCTION seconds_to_duration (
      seconds_in               IN NUMBER,
      full_display_in          IN BOOLEAN,
      include_seconds_in       IN BOOLEAN DEFAULT FALSE,
      trim_leading_zeroes_in   IN BOOLEAN DEFAULT FALSE)
      RETURN VARCHAR2;

   FUNCTION seconds_between (
      from_date_in   IN DATE DEFAULT SYSDATE,
      to_date_in     IN DATE)
      RETURN VARCHAR2;

   FUNCTION time_left_as_duration (date_in IN DATE)
      RETURN VARCHAR2;

   /* For use in SQL. Pass 1 for TRUE, 0 for FALSE */
   FUNCTION seconds_to_duration (
      seconds_in               IN NUMBER,
      full_display_in          IN INTEGER,
      trim_leading_zeroes_in   IN INTEGER DEFAULT 0)
      RETURN VARCHAR2;

   FUNCTION last_daily_quiz_date (
      date_in IN DATE DEFAULT SYSDATE)
      RETURN DATE;
END;
/

CREATE OR REPLACE PACKAGE BODY date_mgr
IS
   FUNCTION q_to_date (q IN VARCHAR2)
      RETURN DATE
   AS
      l_year   CHAR (4);
      l_q      PLS_INTEGER;
   BEGIN

      l_year := SUBSTR (q, 1, 4);
      l_q := SUBSTR (q, 6);

      RETURN TO_DATE (
                   l_year
                || '-'
                || CASE l_q
                      WHEN '1' THEN '01'
                      WHEN '2' THEN '04'
                      WHEN '3' THEN '07'
                      WHEN '4' THEN '10'
                   END
                || '-01',
                'YYYY-MM-DD');
   END q_to_date;

   FUNCTION first_day_in_month (date_in IN DATE)
      RETURN DATE
   IS
   BEGIN
      RETURN TO_DATE (
                '01-' || TO_CHAR (date_in, 'MM-YYYY'),
                'DD-MM-YYYY');
   END first_day_in_month;

   FUNCTION start_of_day (date_in IN DATE DEFAULT SYSDATE)
      RETURN DATE
   IS
   BEGIN
      RETURN TRUNC (date_in) + (1 / (24 * 60 * 60));
   END start_of_day;

   FUNCTION end_of_day (date_in IN DATE DEFAULT SYSDATE)
      RETURN DATE
   IS
   BEGIN
      RETURN TRUNC (date_in) + 1 - (1 / (24 * 60 * 60));
   END end_of_day;

   FUNCTION start_of_week (date_in IN DATE DEFAULT SYSDATE)
      RETURN DATE
   IS
      l_date   DATE;
   BEGIN
      IF TO_CHAR (date_in,
                  'DY',
                  'NLS_DATE_LANGUAGE = AMERICAN') = 'SAT'
      THEN
         l_date := date_in;
      ELSE
         l_date := NEXT_DAY (date_in - 7, 'SATURDAY');
      END IF;

      RETURN TRUNC (l_date) + (1 / (24 * 60 * 60));
   END start_of_week;

   FUNCTION end_of_week (date_in IN DATE DEFAULT SYSDATE)
      RETURN DATE
   IS
      l_date   DATE;
   BEGIN
      IF TO_CHAR (date_in,
                  'DY',
                  'NLS_DATE_LANGUAGE = AMERICAN') = 'FRI'
      THEN
         l_date := date_in;
      ELSE
         l_date := NEXT_DAY (date_in, 'FRIDAY');
      END IF;

      RETURN TRUNC (l_date) + 1 - (1 / (24 * 60 * 60));
   END end_of_week;

   FUNCTION previous_quarter (quarter_in IN VARCHAR2)
      RETURN VARCHAR2
   IS
   BEGIN
      IF quarter_in LIKE '%-1'
      THEN
         RETURN    TO_CHAR (
                        TO_NUMBER (
                           SUBSTR (quarter_in, 1, 4))
                      - 1)
                || '-4';
      ELSE
         RETURN    SUBSTR (quarter_in, 1, 4)
                || '-'
                || TO_CHAR (
                        TO_NUMBER (SUBSTR (quarter_in, 6))
                      - 1);
      END IF;
   END previous_quarter;

   FUNCTION previous_month (month_in IN VARCHAR2)
      RETURN VARCHAR2
   IS
   BEGIN
      IF month_in LIKE '%-01'
      THEN
         RETURN    TO_CHAR (
                        TO_NUMBER (SUBSTR (month_in, 1, 4))
                      - 1)
                || '-12';
      ELSE
         RETURN    SUBSTR (month_in, 1, 4)
                || '-'
                || LPAD (
                      TO_CHAR (
                           TO_NUMBER (SUBSTR (month_in, 6))
                         - 1),
                      2,
                      '0');
      END IF;
   END previous_month;

   FUNCTION dd_to_nth (dd_in IN NUMBER)
      RETURN VARCHAR2
   IS
   BEGIN
      RETURN LOWER (
                TO_CHAR (
                   TO_DATE (
                      '2010-01-' || LPAD (dd_in, 2, '0'),
                      'YYYY-MM-DD'),
                   'DDSPth'));
   END;

   FUNCTION n_to_nth (n_in IN NUMBER)
      RETURN VARCHAR2
   IS
   BEGIN
      RETURN LOWER (
                TO_CHAR (
                   TO_DATE ('1-1-' || n_in, 'dd-mm-yyyy'),
                   'FMYYYYth'));
   END;

   FUNCTION nth_from_n (n_in IN NUMBER)
      RETURN VARCHAR2
   IS
   BEGIN
      RETURN n_to_nth (n_in);
   END;

   FUNCTION date_at_time (date_in IN DATE)
      RETURN VARCHAR2
   IS
   BEGIN
      RETURN    TO_CHAR (date_in, 'FMDD Month YYYY')
             || ' at '
             || TO_CHAR (date_in, 'HH24:MI:SS');
   END;

   FUNCTION q_as_number (date_in IN DATE)
      RETURN NUMBER
   IS
   BEGIN
      RETURN   TO_NUMBER (TO_CHAR (date_in, 'YYYY') * 10)
             + TO_NUMBER (TO_CHAR (date_in, 'Q'));
   END;

   FUNCTION seconds_to_duration (
      seconds_in               IN NUMBER,
      full_display_in          IN BOOLEAN,
      include_seconds_in       IN BOOLEAN DEFAULT FALSE,
      trim_leading_zeroes_in   IN BOOLEAN DEFAULT FALSE)
      RETURN VARCHAR2
   IS
      c_seconds   CONSTANT NUMBER := ROUND (seconds_in);
      c_days      CONSTANT INTEGER
                              := FLOOR (c_seconds / 86400);
      c_hours     CONSTANT INTEGER
                              := FLOOR (c_seconds / 3600);
      c_minutes   CONSTANT INTEGER
         := FLOOR ( (c_seconds - (c_hours * 3600)) / 60);
      c_secs      CONSTANT INTEGER
         := c_seconds - (c_hours * 3600) - (c_minutes * 60);
      l_return             qdb_config.maxvarchar2;
   BEGIN
      IF full_display_in
      THEN
         IF c_days = 0 AND c_hours = 0
         THEN
            l_return :=
               TO_CHAR (
                  TO_DATE (MOD (c_seconds, 86400), 'sssss'),
                  ' mi "mins" ss "secs"');
         ELSE
            l_return :=
                  CASE c_days
                     WHEN 0 THEN NULL
                     WHEN 1 THEN c_days || ' day '
                     ELSE c_days || ' days '
                  END
               || TO_CHAR (
                     TO_DATE (MOD (c_seconds, 86400),
                              'sssss'),
                        'hh24 "hrs" mi "mins"'
                     || CASE
                           WHEN include_seconds_in
                           THEN
                              ' ss "secs"'
                           ELSE
                              NULL
                        END);
         END IF;
      ELSE
         l_return :=
            LTRIM (
               LTRIM (
                     LPAD (c_hours, 4, '0')
                  || ':'
                  || LPAD (c_minutes, 2, '0')
                  || ':'
                  || LPAD (c_secs, 2, '0'),
                  CASE
                     WHEN trim_leading_zeroes_in THEN '0'
                     ELSE 'X'
                  END),
               ':');
      END IF;

      RETURN l_return;
   END seconds_to_duration;

   FUNCTION seconds_to_duration (
      seconds_in               IN NUMBER,
      full_display_in          IN INTEGER,
      trim_leading_zeroes_in   IN INTEGER DEFAULT 0)
      RETURN VARCHAR2
   IS
      bool   BOOLEAN
         := CASE full_display_in
               WHEN 1 THEN TRUE
               ELSE FALSE
            END;
   BEGIN
      RETURN seconds_to_duration (
                seconds_in,
                bool,
                trim_leading_zeroes_in   => CASE trim_leading_zeroes_in
                                              WHEN 1
                                              THEN
                                                 TRUE
                                              ELSE
                                                 FALSE
                                           END);
   END;

   FUNCTION seconds_between (
      from_date_in   IN DATE DEFAULT SYSDATE,
      to_date_in     IN DATE)
      RETURN VARCHAR2
   IS
   BEGIN
      RETURN ROUND (
                TO_NUMBER (
                     (to_date_in - from_date_in)
                   * c_seconds_in_day));
   END seconds_between;

   FUNCTION time_left_as_duration (date_in IN DATE)
      RETURN VARCHAR2
   IS
      l_seconds   NUMBER
                     := seconds_between (SYSDATE, date_in);
   BEGIN
      RETURN seconds_to_duration (
                l_seconds,
                full_display_in      => TRUE,
                include_seconds_in   => CASE
                                          WHEN l_seconds >
                                                  c_seconds_in_day
                                          THEN
                                             FALSE
                                          ELSE
                                             TRUE
                                       END);
   END time_left_as_duration;

   /* Offsets can be N = Now, T = Tomorrow, W = Start of Next week */
   FUNCTION date_from_offset (
      offset_in   IN VARCHAR2,
      date_in     IN DATE DEFAULT NULL)
      RETURN DATE
   IS
      c_date   DATE := NVL (date_in, SYSDATE);
   BEGIN
      RETURN CASE offset_in
                WHEN 'N' THEN c_date
                WHEN 'T' THEN c_date + 1
                WHEN 'W' THEN            /* Next Saturday */
                             NEXT_DAY (c_date + 1, 'SAT')
             END;
   END date_from_offset;
END;
/