declare
   aname varchar2(5);
begin
   begin
      aname := 'Big String';
      dbms_output.put_line(aname);
   exception
      when value_error then
         dbms_output.put_line('Inner block');
   end;

   dbms_output.put_line(sqlcode);
   dbms_output.put_line('What error?');
exception
   when value_error then
      dbms_output.put_line('Outer block');
end;
/

-- An exception raised in the declaration section of a block will always propagate out of that block unhandled.

declare
   aname varchar2(5);
begin
   declare
      aname varchar2(5) := 'Big String';
   begin
      dbms_output.put_line(aname);
   exception
      when value_error then
         dbms_output.put_line('Inner block');
   end;

   dbms_output.put_line(sqlcode);
   dbms_output.put_line('What error?');
exception
   when value_error then
      dbms_output.put_line('Outer block');
end;
/

begin << outer >> declare
   aname varchar2(5);
begin <<inner>>
      declare
         aname   varchar2 (20);
      begin
         outer.aname := 'Big String';
      exception
         when value_error
         then
            raise no_data_found;

         when no_data_found
         then
            dbms_output.put_line ('Inner block');
      end inner;
   exception
      when no_data_found
      then
         dbms_output.put_line ('Outer block');
   end outer;
end;
/

-- Group functions do not raise no_data_found

declare
   v_totsal number;
   v_ename hr.employees.last_name%type;
begin
   /* There are no rows with department_id = -15 */
   select sum (salary) into v_totsal
     from hr.employees 
    where department_id = -15;

   dbms_output.put_line (
      'Total salary: ' || v_totsal);

   select last_name into v_ename
     from hr.employees 
    where salary =
       (select max (salary)
          from emp where department_id = -15);

   dbms_output.put_line (
      'The winner is: ' || v_ename);
exception
   when no_data_found
   then
      dbms_output.put_line ('Outer block');
end;
/

-- Wow. Yes you can use names you might have thought were reserved for Oracle's use. Don't do this.

declare
   pls_integer    varchar2 (1);
   no_data_found  exception;
begin
   select dummy
     into pls_integer
     from dual
    where 1 = 2;

   if pls_integer is null
   then
      raise no_data_found;
   end if;
exception
   when no_data_found
   then
      dbms_output.put_line ('No dummy!');
end;
/

-- You can't use the exception_init pragma with the -1403 error code. Try 100 instead. Or better yet, don't do that. Why would you do that?

DECLARE
   e_bad_date_format   EXCEPTION;
   PRAGMA EXCEPTION_INIT (e_bad_date_format, -1830);
BEGIN
   DBMS_OUTPUT.put_line (TO_DATE ('2010 10 10 44:55:66', 'YYYSS'));
EXCEPTION
   WHEN e_bad_date_format
   THEN
      DBMS_OUTPUT.put_line ('Bad date format');
END;
/

DECLARE
   my_exception   EXCEPTION;
   PRAGMA EXCEPTION_INIT (my_exception, 1830);
BEGIN
   RAISE my_exception;
END;
/

DECLARE
   my_exception   EXCEPTION;
   PRAGMA EXCEPTION_INIT (my_exception, -1403);
BEGIN
   RAISE my_exception;
END;
/

DECLARE
   my_exception   EXCEPTION;
   PRAGMA EXCEPTION_INIT (my_exception, 100);
BEGIN
   RAISE my_exception;
END;
/

-- sqlcode returns 0 when it is called outside of an exception handler.

create or replace show_sqlcode
is
begin
   dbms_output.put_line ('sqlcode in subprogram = ' || sqlcode);
end;
/

begin
   dbms_output.put_line ('sqlcode in execution section = ' || sqlcode);
   show_sqlcode;
   raise value_error;
exception
   when others
   then
      dbms_output.put_line ('sqlcode in exception section = ' || sqlcode);
      show_sqlcode;
end;
/

-- User-defined exceptions always have a sqlcode of 1 unless assigned a different number with the exception_init pragma.

declare
   my_exc exception;
begin
   raise my_exc;
exception
   when others
   then
      dbms_output.put_line ('sqlcode in exception section = ' || sqlcode);
end;
/

declare
   my_exc exception;
   pragma exception_init (my_exc, -20202);
begin
   raise my_exc;
exception
   when others
   then
      dbms_output.put_line ('sqlcode in exception section = ' || sqlcode);
end;
/

-- User-defined exceptions always have an error message of "User-defined exception", unless you raise that exception using raise_application_error.

declare
   my_exc exception;
   pragma exception_init ( my_exc, -20202 );
begin
   raise my_exc;
exception
   when others then
      dbms_output.put_line('sqlerrm in exception section = ' || sqlerrm);
      begin
         raise_application_error(-20202, 'Now you will see this custom error message');
      exception
         when others then
            dbms_output.put_line('sqlerrm after rae = ' || sqlerrm);
      end;
end;
/
-- no_data_found can be raised in a number of ways

drop table my_employees
/

create table my_employees
(
   employee_id   integer
 , last_name     varchar2 (10)
 , salary        number(8)
)
/

begin
   insert into my_employees
        values (100, 'Jobs', 1000000);

   insert into my_employees
        values (200, 'Ellison', 1000000);

   insert into my_employees
        values (300, 'Gates', 1000000);

   commit;
end;
/

create or replace procedure busy_procedure (
   employee_id_in in my_employees.employee_id%type)
is
   l_employee        my_employees%rowtype;
   l_line            varchar2 (1023);
   l_file            utl_file.file_type;
   l_list_of_names   dbms_sql.varchar2s;
begin
   begin
      select *
        into l_employee
        from my_employees e
       where e.employee_id = employee_id_in;

      dbms_output.put_line ('Last Name = ' || l_employee.last_name);
   exception
      when others
      then
         dbms_output.put_line (sqlerrm);
   end;

   begin
      l_file := utl_file.fopen ('TEMP', 'my_employees.txt', 'R');
      utl_file.get_line (l_file, l_line);
      utl_file.get_line (l_file, l_line);
      dbms_output.put_line (l_line);
      utl_file.fclose (l_file);
   exception
      when others
      then
         dbms_output.put_line (sqlerrm);
   end;

   begin
        select last_name
          bulk collect into l_list_of_names
          from my_employees
      order by last_name;

      if l_list_of_names (100) = 'Jobs'
      then
         dbms_output.put_line ('A hard worker!');
      end if;
   exception
      when others
      then
         dbms_output.put_line (sqlerrm);
   end;
end busy_procedure;
/

begin
   busy_procedure (400);
end;
/

-- When you might be raising no_data_found in more than one way...

create or replace procedure who_did_that (
   emp_in in emp.empno%type)
is
   v_ename emp.ename%type;

   line varchar2(1023);
   fid utl_file.file_type;
   eof exception;

   list_of_names dbms_sql.varchar2s;
   undefined_row exception;

   procedure read_file is
   begin
      fid := utl_file.fopen ('c:\temp', 'notme.sql', 'R');
      utl_file.get_line (fid, line);
      utl_file.get_line (fid, line);
   exception 
      when no_data_found
      then
         raise eof;
   end;

   procedure process_collection is
   begin
      if list_of_names (100) > 0
      then
         dbms_out.output_line  ('Positive value at row 100');
      end if;
   exception
      when no_data_found
      then
         raise undefined_row;
   end;
begin
   select ename into v_ename
     from emp
    where empno = emp_in;

   read_file;

   process_collection;

exception
   when no_data_found
   then
      dbms_out.output_line ('implicit query failure only');

   when eof
   then
      dbms_out.output_line ('Read past end of file');

   when undefined_row 
   then
      dbms_out.output_line ('No row in table');

end who_did_that;
/

/* %type and constraints */

create table x (vc varchar2(2))
/

declare
   v_ename x.vc%type;
begin
   v_ename := 'abc'; 
end;
/

create or replace procedure set_it (p_vc in out x.vc%type)
is
begin
   p_vc := 'abc';
end;
/

declare
   v_ename varchar2(100);
begin
   set_it (v_ename); 
end;
/

-- When you use save exceptions in forall, if at least one exception was raised, the forall statement will raise a ORA-21438 error after all iterations are attempted.

declare
   type emps_t is table of my_employees%rowtype index by pls_integer;
   l_emps emps_t;
begin
   select * bulk collect into l_emps from my_employees;

   forall indx in 1 .. l_emps.count
      update my_employees set name = name || name where id = l_emp(indx).id;
end;
/ declare
   type emps_t is
      table of my_employees%rowtype index by pls_integer;
   l_emps emps_t;
begin select * bulk collect into l_emps from my_employees ; forall indx in 1 . . l_emps . count save exceptions update my_employees set name = name | | name where id = l_emp ( indx ) . id ; end ;
/

-- When you use log errors with a SQL statement, no error will be raised once the statement completes (no matter how many errors might have occurred at the row level).
 begin dbms_errlog . create_error_log ( dml_table_name = > 'MY_EMPLOYEES' ) ; end ;
/ begin update my_employees set name = name | | name ; end ;
/ begin update my_employees set name = name | | name log errors reject limit unlimited ; end ;
/