CREATE OR REPLACE PROCEDURE prepare_raise_value (
   raise_in       IN     employees.salary%TYPE,
   employee_in    IN     employees.employee_id%TYPE,
   salary_inout   IN OUT employees.salary%TYPE)
IS
BEGIN
   /*
      Placeholder for program with VERY complicated logic, 
      justifying the running of this SQL statement inside a PL/SQL loop.
   */
   NULL;
END;
/

CREATE OR REPLACE PROCEDURE raise_across_dept1 (
   dept_in    IN employees.department_id%TYPE,
   raise_in   IN employees.salary%TYPE) 
IS
   /* 
     One step with cursor FOR loop:
     The kind of code that is likely to cause
     Snapshot too old and Rollback segment too small errors
   */

   CURSOR emp_cur
   IS
      SELECT employee_id, salary
        FROM employees
       WHERE department_id = dept_in;
BEGIN
   FOR rec IN emp_cur
   LOOP
      prepare_raise_value (raise_in, rec.employee_id, rec.salary);

      UPDATE employees
         SET salary = rec.salary
       WHERE employee_id = rec.employee_id;
   END LOOP;
END raise_across_dept1;
/

CREATE OR REPLACE PROCEDURE raise_across_dept (
   dept_in           IN employees.department_id%TYPE,
   raise_in          IN employees.salary%TYPE,
   commit_after_in   IN PLS_INTEGER)
IS
  /*
     Traditional incremental commit logic.
     This will not help at all, though, with Snapshot too old errors.
   */
   l_counter   PLS_INTEGER;

   CURSOR emp_cur
   IS
      SELECT employee_id, salary
        FROM employees
       WHERE department_id = dept_in;
BEGIN
   l_counter := 1;

   FOR rec IN emp_cur
   LOOP
      prepare_raise_value (raise_in, rec.employee_id, rec.salary);

      UPDATE employees
         SET salary = rec.salary
       WHERE employee_id = rec.employee_id;

      IF l_counter >= commit_after_in
      THEN
         COMMIT;
         l_counter := 1;
      ELSE
         l_counter := l_counter + 1;
      END IF;
   END LOOP;

   COMMIT;
END;
/

CREATE OR REPLACE PROCEDURE raise_across_dept (
   dept_in           IN employees.department_id%TYPE,
   raise_in          IN employees.salary%TYPE,
   commit_after_in   IN PLS_INTEGER)
IS
  /*
    Use of BULK COLLECT could bypass Snapshot too old problem
    Use of FORALL could hit the Rollback segment too small problem FASTER
  */
   l_counter      PLS_INTEGER;

   TYPE employee_tt IS TABLE OF employees.employee_id%TYPE
      INDEX BY BINARY_INTEGER;

   employee_ids   employee_tt;

   TYPE salary_tt IS TABLE OF employees.salary%TYPE
      INDEX BY BINARY_INTEGER;

   salaries       salary_tt;
BEGIN
   SELECT employee_id, salary
     BULK COLLECT INTO employee_ids, salaries
     FROM employees
    WHERE department_id = dept_in;

   FOR indx IN employee_ids.FIRST .. employee_ids.LAST
   LOOP
      prepare_raise_value (raise_in, employee_ids (indx), salaries (indx));
   END LOOP;

   FORALL indx IN employee_ids.FIRST .. employee_ids.LAST
      UPDATE employees
         SET salary = salaries (indx)
       WHERE employee_id = employee_ids (indx);

   COMMIT;
END;
/

REM Incremental commit processing with FORALL

CREATE OR REPLACE PROCEDURE raise_across_dept (
   dept_in           IN employees.department_id%TYPE,
   raise_in          IN employees.salary%TYPE,
   limit_bulk_in     IN PLS_INTEGER DEFAULT 10000,
   commit_after_in   IN PLS_INTEGER DEFAULT 1000)
IS
   l_last         PLS_INTEGER;
   l_start        PLS_INTEGER;
   l_end          PLS_INTEGER;

   type emp_info_rt is record (employees.employee_id%TYPE, employees.salary%TYPE);

   TYPE emp_info_tt IS TABLE OF emp_info_rt
      INDEX BY PLS_INTEGER;

   l_emp_info emp_info_tt;

   CURSOR emps_in_dept_cur
   IS
      SELECT employee_id, salary
        FROM employee
       WHERE department_id = dept_in;
BEGIN
   OPEN emps_in_dept_cur;

   LOOP
      FETCH emps_in_dept_cur
         BULK COLLECT INTO l_emp_info
         LIMIT limit_bulk_in;

      EXIT WHEN l_emp_info.COUNT = 0;

      l_last := l_emp_info.COUNT;
      l_start := 1;

      FOR indx IN l_start .. l_last
      LOOP
         prepare_raise_value (raise_in, l_emp_info.employee_id (indx), l_emp_info.salary (indx));
      END LOOP;

      /* WARNING: I have never actually run this code! */
      LOOP
         EXIT WHEN l_start > l_last;
         l_end := LEAST (l_start + commit_after_in, l_last);

         BEGIN
            FORALL indx IN l_start .. l_end SAVE EXCEPTIONS
               UPDATE employee
                  SET salary = l_emp_info.salary (indx)
                WHERE employee_id = l_emp_info.employee_id (indx);
         EXCEPTION
            WHEN errpkg.bulk_errors
            THEN
               /* Write errors out to your REAL error log. 
                  Here I simply display the information. */
               DBMS_OUTPUT.put_line (DBMS_UTILITY.format_error_stack);
               DBMS_OUTPUT.put_line ('Updated ' || SQL%ROWCOUNT || ' rows.');

               FOR indx IN 1 .. SQL%BULK_EXCEPTIONS.COUNT
               LOOP
                  DBMS_OUTPUT.put_line (
                        'Error '
                     || indx
                     || ' occurred on index '
                     || SQL%BULK_EXCEPTIONS (indx).ERROR_INDEX
                     || ' attempting to update name to "'
                     || enames_with_errors (
                           SQL%BULK_EXCEPTIONS (indx).ERROR_INDEX)
                     || '"');
                  DBMS_OUTPUT.put_line (
                        'Oracle error is '
                     || SQLERRM (-1 * SQL%BULK_EXCEPTIONS (indx).ERROR_CODE));
               END LOOP;
         END;

         COMMIT;
         l_start := l_end + 1;
      END LOOP;

      COMMIT;
   END LOOP;
END;
/