create or replace PACKAGE BODY ora_logger
IS
   TYPE t_vc2_vc2_array IS TABLE OF VARCHAR2 (100)
      INDEX BY VARCHAR2 (10);

   g_date_formats                t_vc2_vc2_array;
   g_period_formats              t_vc2_vc2_array;
   g_last_timing                 NUMBER := NULL;
   g_max_log_entries             INTEGER;

   /* 3.0.9 Cut down on overhead of checking tracing */
   g_check_trace_secs   CONSTANT INTEGER := 20;
   g_last_trace_check_on         DATE;
   g_last_trace_check            BOOLEAN;

   /* Trace */
   TYPE trace_rt IS RECORD
   (
      tracing         BOOLEAN DEFAULT FALSE,
      trace_context   VARCHAR2(32767),
      callstack       VARCHAR2(32767),
      trace_target    VARCHAR2(32767)
   );

   g_trace                       trace_rt;

   PROCEDURE display_header (header_in             IN VARCHAR2,
                             length_in             IN PLS_INTEGER DEFAULT 80,
                             border_character_in   IN VARCHAR2 DEFAULT '=',
                             centered_in           IN BOOLEAN DEFAULT FALSE,
                             to_log_in             IN BOOLEAN DEFAULT FALSE,
                             log_context_in        IN VARCHAR2 DEFAULT NULL,
                             to_collection_in      IN BOOLEAN DEFAULT FALSE,
                             indent_in             IN INTEGER DEFAULT 0,
                             prefix_in             IN VARCHAR2 DEFAULT '|')
   IS
      FUNCTION padit (str VARCHAR2)
         RETURN VARCHAR2
      IS
      BEGIN
         RETURN    CASE WHEN prefix_in IS NULL THEN NULL ELSE '| ' END
                || LPAD (str, LENGTH (str) + indent_in, ' ');
      END;

      PROCEDURE pl (str VARCHAR2)
      IS
      BEGIN
         IF to_log_in
         THEN
            trace_activity_force (log_context_in, padit (str));
         ELSIF to_collection_in
         THEN
            g_output (g_output.COUNT + 1) := padit (str);
         ELSE
            DBMS_OUTPUT.put_line (padit (str));
         END IF;
      END;
   BEGIN
      pl (RPAD (border_character_in, length_in, border_character_in));
      pl (
         CASE
            WHEN centered_in
            THEN
                  '|'
               || LPAD (' ', (length_in - LENGTH (header_in)) / 2)
               || header_in
            ELSE
               header_in
         END);
      pl (RPAD (border_character_in, length_in, border_character_in));
   END display_header;

   PROCEDURE trace_and_raise (context_in IN VARCHAR2, text_in IN VARCHAR2)
   IS
   BEGIN
      trace_activity_force (
         context_in,
            text_in
         || CHR (10)
         || DBMS_UTILITY.format_error_stack ()
         || CHR (10)
         || DBMS_UTILITY.format_error_backtrace);
      raise_application_error (
         -20000,
            context_in
         || ': '
         || text_in
         || CHR (10)
         || DBMS_UTILITY.format_error_stack ()
         || CHR (10)
         || DBMS_UTILITY.format_error_backtrace);
   END;

   PROCEDURE trace_sys_contexts (context_in IN VARCHAR2 DEFAULT NULL)
   IS
      c_context   CONSTANT VARCHAR2(32767) := 'sys_context';

      PROCEDURE trace_context (NAME_IN IN VARCHAR2)
      IS
      BEGIN
         trace_activity (c_context || ' ' || NAME_IN, 'NOTHING');
      END;
   BEGIN
      IF trace_is_on
      THEN
         trace_activity (c_context, context_in);
         trace_context ('current_domain_id');
         trace_context ('current_competition_id');
         trace_context ('current_user_id');
         trace_context ('current_company_name');
         trace_context ('current_country');
      END IF;
   END trace_sys_contexts;

   /* System log */

   PROCEDURE syslog (logtype_in   IN     VARCHAR2,
                     action_in    IN     VARCHAR2,
                     title_in     IN     VARCHAR2,
                     details_in   IN     VARCHAR2,
                     log_id_out      OUT NUMBER)
   IS
      PRAGMA AUTONOMOUS_TRANSACTION;
      l_report        CLOB;
   BEGIN
         INSERT INTO system_log (logtype,
                                     action,
                                     title,
                                     details,
                                     callstack,
                                     errorstack,
                                     errortrace)
                 VALUES (
                           logtype_in,
                           action_in,
                           title_in,
                           details_in,
                           DBMS_UTILITY.format_call_stack, 
                           DBMS_UTILITY.format_error_stack, 
                           DBMS_UTILITY.format_error_backtrace)
           RETURNING system_log_id
                INTO log_id_out;

         COMMIT;
   EXCEPTION
      WHEN OTHERS
      THEN
         DECLARE
            l_code   INTEGER := SQLCODE;
            l_errm   VARCHAR2(32767)
                        := DBMS_UTILITY.format_error_stack ();
         BEGIN
            ROLLBACK;
            trace_activity_force (
               'SYSLOG FAILURE',
                  'Unable to write to system log. Error code/message: '
               || l_code
               || '/'
               || l_errm);
         END;
   END syslog;

   PROCEDURE syslog (logtype_in   IN VARCHAR2,
                     action_in    IN VARCHAR2,
                     title_in     IN VARCHAR2,
                     details_in   IN VARCHAR2)
   IS
      l_id   INTEGER;
   BEGIN
      syslog (logtype_in   => logtype_in,
              action_in    => action_in,
              title_in     => title_in,
              details_in   => details_in,
              log_id_out   => l_id);
   END syslog;

   /* 2017-01-10 Clob overloading for full details */
   PROCEDURE syslog (logtype_in        IN VARCHAR2,
                     action_in         IN VARCHAR2,
                     title_in          IN VARCHAR2,
                     details_in        IN VARCHAR2,
                     full_details_in   IN CLOB)
   IS
      PRAGMA AUTONOMOUS_TRANSACTION;
      l_app_session   NUMBER;
      l_app_id        NUMBER;
      l_page_id       NUMBER;
      l_id            INTEGER;

      l_count         INTEGER;
   BEGIN
      /* 2.5.3 If > 100,000 rows do not write. Something has gone haywire. */
      SELECT COUNT (*) INTO l_count FROM system_log;

      IF l_count < 100000
      THEN
         l_app_session := v ('APP_SESSION');
         l_app_id := v ('APP_ID');
         l_page_id := v ('APP_PAGE_ID');

         INSERT INTO system_log (logtype,
                                     action,
                                     title,
                                     details,
                                     full_details,
                                     callstack,
                                     errorstack,
                                     errortrace,
                                     app_id,
                                     session_id,
                                     page_id)
                 VALUES (
                           logtype_in,
                           SUBSTR (action_in, 1, 4000),
                           SUBSTR (title_in, 1, 4000),
                              SUBSTR (details_in, 1, 3000)
                           || CASE
                                 WHEN LENGTH (details_in) > 3000
                                 THEN
                                    '...see FULL_DETAILS for rest of text'
                              END,
                           details_in,
                           SUBSTR (DBMS_UTILITY.format_call_stack, 1, 4000),
                           SUBSTR (DBMS_UTILITY.format_error_stack, 1, 4000),
                           SUBSTR (DBMS_UTILITY.format_error_backtrace,
                                   1,
                                   4000),
                           l_app_id,
                           l_app_session,
                           l_page_id)
           RETURNING system_log_id
                INTO l_id;

         COMMIT;
      END IF;
   EXCEPTION
      WHEN OTHERS
      THEN
         DECLARE
            l_code   INTEGER := SQLCODE;
            l_errm   VARCHAR2(32767)
                        := DBMS_UTILITY.format_error_stack ();
         BEGIN
            ROLLBACK;
            trace_activity_force (
               'SYSLOG FAILURE',
                  'Unable to write to system log. Error code/message: '
               || l_code
               || '/'
               || l_errm);
         END;
   END syslog;

   PROCEDURE syslog_error (action_in         IN     VARCHAR2,
                           title_in          IN     VARCHAR2,
                           details_in        IN     VARCHAR2,
                           raise_error_in    IN     BOOLEAN DEFAULT FALSE,
                           message_only_in   IN     BOOLEAN DEFAULT FALSE,
                           user_message_in   IN     VARCHAR2 DEFAULT NULL,
                           log_id_out           OUT NUMBER)
   IS
   BEGIN
      IF raise_error_in AND SQLCODE = c_reraise_errnum
      THEN
         raise_application_error (
            c_reraise_errnum,
            LTRIM (SUBSTR (SQLERRM, INSTR (SQLERRM, ':') + 1)));
      ELSE
         syslog ('ERROR',
                 action_in,
                 title_in,
                 details_in,
                 log_id_out   => log_id_out);

         IF raise_error_in
         THEN
            /* 3.1 Allow override of message */
            raise_application_error (
               c_reraise_errnum,
               CASE
                  WHEN user_message_in IS NOT NULL
                  THEN
                     user_message_in
                  ELSE
                        action_in
                     || ': '
                     || title_in
                     || ' '
                     || details_in
                     || CASE
                           WHEN message_only_in
                           THEN
                              NULL
                           ELSE
                                 ' '
                              || sys.DBMS_UTILITY.format_error_stack
                              || ' '
                              || sys.DBMS_UTILITY.format_error_backtrace
                        END
               END);
         END IF;
      END IF;
   END;

   PROCEDURE syslog_error (action_in         IN VARCHAR2,
                           title_in          IN VARCHAR2,
                           details_in        IN VARCHAR2,
                           raise_error_in    IN BOOLEAN DEFAULT FALSE,
                           message_only_in   IN BOOLEAN DEFAULT FALSE,
                           user_message_in   IN VARCHAR2 DEFAULT NULL)
   IS
      l_id   INTEGER;
   BEGIN
      syslog_error (action_in         => action_in,
                    title_in          => title_in,
                    details_in        => details_in,
                    raise_error_in    => raise_error_in,
                    message_only_in   => message_only_in,
                    user_message_in   => user_message_in,
                    log_id_out        => l_id);
   END;

   PROCEDURE syslog_information (action_in    IN VARCHAR2,
                                 title_in     IN VARCHAR2,
                                 details_in   IN VARCHAR2 DEFAULT NULL)
   IS
   BEGIN
      syslog ('INFORMATION',
              action_in,
              title_in,
              details_in);
   END;

   PROCEDURE syslog_warning (action_in    IN VARCHAR2,
                             title_in     IN VARCHAR2,
                             details_in   IN VARCHAR2)
   IS
   BEGIN
      syslog ('WARNING',
              action_in,
              title_in,
              details_in);
   END;

   PROCEDURE syslog_assert (condition_in IN BOOLEAN, text_in IN VARCHAR2)
   IS
   BEGIN
      IF condition_in IS NULL OR NOT condition_in
      THEN
         syslog_error ('ASSERTION-FAILURE', text_in, NULL);
         raise_application_error (-20000, text_in);
      END IF;
   END syslog_assert;
   
   /* By default send report for previous day, up to sysdate,
      to the support email list.
   */

   PROCEDURE send_syslog_report (from_in      IN DATE DEFAULT NULL,
                                 to_in        IN DATE DEFAULT NULL,
                                 send_to_in   IN VARCHAR2 DEFAULT NULL)
   IS
      l_start          DATE := NVL (from_in, TRUNC (SYSDATE - 1));
      l_end            DATE := NVL (to_in, SYSDATE);
      l_empty          BOOLEAN;
      l_count          INTEGER;
      l_job_running    VARCHAR2 (1);
      l_queue_active   VARCHAR2 (1);
      l_validation     VARCHAR2(32767);
      l_report         CLOB
         := 'TRACE ' || CASE WHEN trace_is_on () THEN 'ON' ELSE 'OFF' END;

      PROCEDURE build_report (empty_out OUT BOOLEAN)
      IS
         PROCEDURE add_string (string_in IN VARCHAR2)
         IS
         BEGIN
            IF string_in IS NOT NULL
            THEN
               l_report := l_report || '<br/><br/>' || string_in;
            END IF;
         END;
      BEGIN
         empty_out := TRUE;

         /* Only report on errors */

         FOR rec
            IN (     SELECT *
                       FROM system_log sl
                      WHERE     sl.created_on BETWEEN l_start AND l_end
                            AND sl.action NOT IN ('REMINDERS_SENT')
                            AND sl.logtype = 'ERROR'
                   ORDER BY sl.created_on DESC
                FETCH FIRST 100 ROWS ONLY)
         LOOP
            empty_out := FALSE;
            add_string (
                  rec.logtype
               || ': '
               || rec.action
               || ' | '
               || rec.title
               || ' | Created on/by: '
               || TO_CHAR (rec.created_on, 'YYYY-MM-DD HH24:MI:SS')
               || '/'
               || rec.created_by);

            IF rec.details IS NOT NULL
            THEN
               FOR indx IN 1 ..
                           NVL (TRUNC (LENGTH (rec.details) / 1000) + 1, 0)
               LOOP
                  add_string (
                     SUBSTR (rec.details, (indx - 1) * 1000 + 1, 1000));
               END LOOP;
            END IF;
         END LOOP;

         add_string (l_validation);
         
         SELECT COUNT (*) INTO l_count FROM execution_trace;

         IF l_count > 1000
         THEN
            add_string ('Trace Log contains ' || l_count || ' rows.');
         END IF;
      EXCEPTION
         WHEN VALUE_ERROR
         THEN
            NULL;
      END;
   BEGIN
      build_report (l_empty);

      IF NOT l_empty
      THEN
         apex_mail.send (
            from_email_address_in   => 'noreply@oracle.com',
            to_email_address_in     => NVL (send_to_in,
                                            support_emails()),
            subject_in              =>    'Dev Gym SysLog Report for '
                                       || TO_CHAR (l_start,
                                                   'YYYY-MM-DD HH24:MI')
                                       || ' - '
                                       || TO_CHAR (l_end, 'YYYY-MM-DD HH24:MI'),
            html_in                 => l_report,
            text_in                 => NULL);
      END IF;
   END send_syslog_report;

   PROCEDURE send_syslog_report_last_hour
   IS
   BEGIN
      send_syslog_report (SYSDATE - 1 / 24, SYSDATE);
   END;

   /* Trace functionality */

   PROCEDURE turn_on_trace (
      context_filter_in       IN VARCHAR2 DEFAULT NULL,
      this_session_only_in    IN BOOLEAN DEFAULT FALSE,
      callstack_contains_in   IN VARCHAR2 DEFAULT NULL,
      send_trace_to_in        IN VARCHAR2 DEFAULT c_to_table,
      max_log_entries_in      IN INTEGER DEFAULT NULL)
   IS
      PRAGMA AUTONOMOUS_TRANSACTION;
   BEGIN
      IF this_session_only_in
      THEN
         g_trace.callstack := UPPER (callstack_contains_in);
         g_trace.trace_target := send_trace_to_in;
      ELSE
         /* 2016-10-19 SF Fix WHEN to avoid endless string of "%" for filter */
         UPDATE app_config
            SET trace_setting = 'Y',
                trace_filter =
                   CASE
                      WHEN NVL (context_filter_in, '%') = '%' THEN '%'
                      ELSE '%' || UPPER (context_filter_in) || '%'
                   END,
                trace_callstack = UPPER (callstack_contains_in),
                trace_target = send_trace_to_in;

         COMMIT;
      END IF;

      g_max_log_entries := max_log_entries_in;

      /* Always set global flag, so that multiple calls in the same
        server call can use this and not select from the table. */
      g_trace.tracing := TRUE;
      g_trace.trace_context := '%' || UPPER (context_filter_in) || '%';
   END;

   PROCEDURE turn_off_trace
   IS
      PRAGMA AUTONOMOUS_TRANSACTION;
   BEGIN
      g_trace.tracing := FALSE;
      g_max_log_entries := NULL;

      UPDATE app_config
         SET trace_setting = 'N',
             trace_filter = '%',
             trace_user_id = NULL;

      COMMIT;
   END;

   FUNCTION trace_is_on (context_filter_in IN VARCHAR2 DEFAULT NULL)
      RETURN BOOLEAN
   IS
      l_app_config   app_config%ROWTYPE;
      l_trace        CHAR (1);
      l_filter       VARCHAR2 (100);
      l_return       BOOLEAN DEFAULT FALSE;
   BEGIN
      /* If this flag is set, then we are done. */
      IF g_trace.tracing
      THEN
         l_return :=
               context_filter_in IS NULL
            OR UPPER (context_filter_in) LIKE g_trace.trace_context;
      /* 3.0.9 Don't check the table for 20 secs since last check */
      /* 3.1 skip this - seems to be causing issues likely with stateless execution
      ELSIF (SYSDATE - g_last_trace_check_on) * (24 * 60 * 60) <= g_check_trace_secs
      THEN
         l_return := g_last_trace_check;*/
      ELSE
         SELECT *
           INTO l_app_config
           FROM app_config
          WHERE ROWNUM < 2;

         l_app_config.trace_filter := NVL (l_app_config.trace_filter, '%');

         IF l_app_config.trace_setting = 'Y'
         THEN
            IF l_app_config.trace_user_id IS NULL
            THEN
               l_return := TRUE;
            ELSE
               l_return := l_app_config.trace_user_id = v ('AI_USER_ID');
            END IF;
         END IF;

         IF l_return AND context_filter_in IS NOT NULL
         THEN
            l_return := UPPER (context_filter_in) LIKE l_app_config.trace_filter;
         END IF;
      END IF;

      --g_last_trace_check := l_return;
      --g_last_trace_check_on := SYSDATE;

      RETURN l_return;
   END trace_is_on;

   PROCEDURE check_too_many
   IS
      l_count   INTEGER;
      l_max     INTEGER;
   BEGIN
      SELECT COUNT (*) INTO l_count FROM execution_trace;

      SELECT max_log_rows INTO l_max FROM app_config;

      IF l_count > l_max
      THEN
         UPDATE app_config
            SET ignore_logging = 'Y';
      END IF;
   END;

   PROCEDURE internal_trace (context_in    IN VARCHAR2,
                             data_in       IN VARCHAR2,
                             override_in   IN BOOLEAN)
   IS
      PRAGMA AUTONOMOUS_TRANSACTION;
      c_callstack   CONSTANT VARCHAR2(32767)
                                := DBMS_UTILITY.format_call_stack ;
      l_app_config           app_config%ROWTYPE;

      l_trace_it             BOOLEAN := override_in OR trace_is_on (context_in);
      l_text                 VARCHAR2(32767)
         := CASE
               WHEN LENGTH (data_in) > 3000 THEN 'See FULL_TEXT'
               ELSE data_in
            END;
      l_callstack            VARCHAR2(32767)
         := CASE
               WHEN LENGTH (c_callstack) > 3000 THEN 'See FULL_TEXT'
               ELSE c_callstack
            END;
      c_context     CONSTANT execution_trace.text%TYPE := SUBSTR (context_in, 1, 4000);
   BEGIN
      IF l_trace_it
      THEN
         SELECT *
           INTO l_app_config
           FROM app_config
          WHERE ROWNUM < 2;

         IF l_app_config.trace_callstack IS NOT NULL
         THEN
            l_trace_it := INSTR (c_callstack, l_app_config.trace_callstack) > 0;
         END IF;
      END IF;

      IF l_trace_it
      THEN
         IF g_max_log_entries IS NOT NULL
         THEN
            DECLARE
               l_count   INTEGER;
            BEGIN
               SELECT COUNT (*) INTO l_count FROM execution_trace;

               IF l_count >= g_max_log_entries
               THEN
                  raise_application_error (
                     -20000,
                     'Exceeded maximum log entries, terminating runaway process.');
               END IF;
            END;
         END IF;

         CASE
            WHEN    l_app_config.trace_target = c_to_screen
                 OR g_trace.trace_target = c_to_screen
            THEN
               putline (context_in || ': ' || data_in);
            /* 2016-11-06 SF NO longer run out of tablespace */
            WHEN l_app_config.ignore_logging = 'Y'
            THEN
               NULL;
            ELSE
               INSERT INTO execution_trace (context,
                                    full_text,
                                    full_callstack,
                                    text,
                                    callstack)
                    VALUES (c_context,
                            data_in,
                            c_callstack,
                            l_text,
                            l_callstack);

               check_too_many;

               COMMIT;
         END CASE;
      END IF;
   END;

   PROCEDURE internal_trace (context_in    IN VARCHAR2,
                             data_in       IN CLOB,
                             override_in   IN BOOLEAN)
   IS
      PRAGMA AUTONOMOUS_TRANSACTION;
      c_app_config   CONSTANT app_config%ROWTYPE
                                 := app_config_row () ;
      l_trace_it              BOOLEAN
                                 := override_in OR trace_is_on (context_in);
      c_context      CONSTANT execution_trace.text%TYPE := SUBSTR (context_in, 1, 4000);
   BEGIN
      IF NOT l_trace_it AND c_app_config.trace_callstack IS NOT NULL
      THEN
         l_trace_it :=
            INSTR (DBMS_UTILITY.format_call_stack,
                   c_app_config.trace_callstack) > 0;
      END IF;

      IF l_trace_it
      THEN
         IF g_max_log_entries IS NOT NULL
         THEN
            DECLARE
               l_count   INTEGER;
            BEGIN
               SELECT COUNT (*) INTO l_count FROM execution_trace;

               IF l_count >= g_max_log_entries
               THEN
                  raise_application_error (
                     -20000,
                     'Exceeded maximum log entries, terminating runaway process.');
               END IF;
            END;
         END IF;

         CASE
            WHEN c_app_config.trace_target = c_to_screen
            THEN
               putline (context_in || ': ' || data_in);
            /* 2016-11-06 SF NO longer run out of tablespace */
            WHEN c_app_config.ignore_logging = 'Y'
            THEN
               NULL;
            ELSE
               INSERT INTO execution_trace (context,
                                    full_text,
                                    full_callstack,
                                    text,
                                    callstack)
                       VALUES (
                                 c_context,
                                 data_in,
                                 DBMS_UTILITY.format_call_stack,
                                 'See FULL_TEXT',
                                 CASE
                                    WHEN LENGTH (
                                            DBMS_UTILITY.format_call_stack) >
                                            3000
                                    THEN
                                       'See FULL_CALLSTACK'
                                    ELSE
                                       DBMS_UTILITY.format_call_stack
                                 END);

               check_too_many;
               COMMIT;
         END CASE;
      END IF;
   END;

   PROCEDURE trace_activity (context_in IN VARCHAR2, data_in IN VARCHAR2)
   IS
   BEGIN
      internal_trace (context_in, data_in, override_in => FALSE);
   END;

   PROCEDURE trace_activity (
      context_in   IN VARCHAR2,
      data_in      IN DATE,
      format_in    IN VARCHAR2 DEFAULT 'YYYY-MM-DD HH24:MI:SS')
   IS
   BEGIN
      internal_trace (context_in,
                      TO_CHAR (data_in, format_in),
                      override_in   => FALSE);
   END;

   PROCEDURE trace_activity (context_in IN VARCHAR2, data_in IN CLOB)
   IS
   BEGIN
      internal_trace (context_in, data_in, override_in => FALSE);
   END;

   PROCEDURE trace_activity (context_in IN VARCHAR2, data_in IN BOOLEAN)
   IS
   BEGIN
      internal_trace (context_in,
                      bool2vc (data_in, tf_in => TRUE),
                      override_in   => FALSE);
   END;

   /* Use this inside expressions and SQL */

   FUNCTION trace_activity (context_in IN VARCHAR2, data_in IN VARCHAR2)
      RETURN PLS_INTEGER
   IS
   BEGIN
      trace_activity (context_in, data_in);
      RETURN 1;
   END;

   FUNCTION trace_activity_force (context_in IN VARCHAR2, data_in IN VARCHAR2)
      RETURN PLS_INTEGER
   IS
   BEGIN
      trace_activity_force (context_in, data_in);
      RETURN 1;
   END;

   PROCEDURE trace_activity_if (condition_in   IN BOOLEAN,
                                context_in     IN VARCHAR2,
                                data_in        IN VARCHAR2)
   IS
   BEGIN
      IF condition_in
      THEN
         trace_activity_force (context_in, data_in);
      END IF;
   END;

   PROCEDURE trace_activity_if (condition_in   IN BOOLEAN,
                                context_in     IN BOOLEAN,
                                data_in        IN VARCHAR2)
   IS
   BEGIN
      IF condition_in
      THEN
         trace_activity_force (bool2vc (context_in), data_in);
      END IF;
   END;

   FUNCTION trace_activity_if_sql (equals_1_in   IN PLS_INTEGER,
                                   context_in    IN VARCHAR2,
                                   data_in       IN VARCHAR2)
      RETURN PLS_INTEGER
   IS
   BEGIN
      trace_activity_if (CASE equals_1_in WHEN 1 THEN TRUE ELSE FALSE END,
                         context_in,
                         data_in);
      RETURN 1;
   END;

   PROCEDURE trace_activity_force (context_in IN VARCHAR2, data_in IN VARCHAR2)
   IS
   BEGIN
      internal_trace (context_in, data_in, override_in => TRUE);
   END;

   PROCEDURE trace_activity_force (context_in IN VARCHAR2, data_in IN CLOB)
   IS
   BEGIN
      internal_trace (context_in, data_in, override_in => TRUE);
   END;

   PROCEDURE trace_activity_force (context_in IN VARCHAR2, data_in IN BOOLEAN)
   IS
   BEGIN
      internal_trace (context_in,
                      bool2vc (data_in, tf_in => TRUE),
                      override_in   => TRUE);
   END;

   PROCEDURE taf (data_in IN VARCHAR2)
   IS
   BEGIN
      trace_activity_force ('TAF', data_in);
   END;

   PROCEDURE taf (data_in IN CLOB)
   IS
   BEGIN
      trace_activity_force ('TAF', data_in);
   END;

   PROCEDURE taf (data_in IN BOOLEAN)
   IS
   BEGIN
      trace_activity_force ('TAF', data_in);
   END;

   PROCEDURE clear_trace_entries (before_in IN DATE DEFAULT NULL)
   IS
      PRAGMA AUTONOMOUS_TRANSACTION;
   BEGIN
      IF before_in IS NULL
      THEN
         EXECUTE IMMEDIATE 'truncate table execution_trace';
      ELSE
         DELETE FROM execution_trace
               WHERE created_on <= NVL (before_in, SYSDATE);
      END IF;

      UPDATE app_config
         SET ignore_logging = 'N';

      COMMIT;
   END;

   PROCEDURE write_trace_to_file (dir_in IN VARCHAR2, file_in IN VARCHAR2)
   IS
      l_fhandler   UTL_FILE.file_type;
   BEGIN
      l_fhandler :=
         UTL_FILE.fopen (dir_in,
                         file_in,
                         'W',
                         32767);

      FOR rec IN (  SELECT *
                      FROM execution_trace
                  ORDER BY log_id)
      LOOP
         UTL_FILE.put_line (l_fhandler, rec.text);
      END LOOP;

      UTL_FILE.fclose (l_fhandler);
   END;

   /* End Trace Functionality */

   /* Substitutes for DBMS_OUTPUT.PUT_LINE */

   PROCEDURE pl (string_in          IN VARCHAR2,
                 show_null_in       IN BOOLEAN DEFAULT TRUE,
                 to_log_in          IN BOOLEAN DEFAULT FALSE,
                 log_context_in     IN VARCHAR2 DEFAULT NULL,
                 to_collection_in   IN BOOLEAN DEFAULT FALSE,
                 indent_in          IN INTEGER DEFAULT 0,
                 prefix_in          IN VARCHAR2 DEFAULT '|')
   IS
      FUNCTION padit (str VARCHAR2, indent_in IN INTEGER)
         RETURN VARCHAR2
      IS
      BEGIN
         RETURN    CASE WHEN prefix_in IS NULL THEN NULL ELSE '| ' END
                || LPAD (str, LENGTH (str) + indent_in, ' ');
      END;
   BEGIN
      IF show_null_in OR string_in IS NOT NULL
      THEN
         IF to_log_in
         THEN
            trace_activity_force (log_context_in, padit (string_in, indent_in));
         ELSIF to_collection_in
         THEN
            g_output (g_output.COUNT + 1) := padit (string_in, indent_in);
         ELSE
            DBMS_OUTPUT.put_line (padit (string_in, indent_in));
         END IF;
      END IF;
   END;

   PROCEDURE putline (data_in IN VARCHAR2)
   IS
   BEGIN
      DBMS_OUTPUT.put_line (data_in);
   END;

   PROCEDURE putline (data_in     IN DATE,
                      format_in   IN VARCHAR2 DEFAULT 'YYYY-MM-DD HH24:MI:SS')
   IS
   BEGIN
      DBMS_OUTPUT.put_line (TO_CHAR (data_in, format_in));
   END;

   PROCEDURE putline (data_in IN BOOLEAN)
   IS
   BEGIN
      DBMS_OUTPUT.put_line (bool2vc (data_in, tf_in => TRUE));
   END;

   PROCEDURE putline (title_in IN VARCHAR2, data_in IN BOOLEAN)
   IS
   BEGIN
      DBMS_OUTPUT.put_line (
         title_in || ' - ' || bool2vc (data_in, tf_in => TRUE));
   END;

   PROCEDURE putline (clob_in            IN CLOB,
                      append_in          IN BOOLEAN DEFAULT TRUE,
                      preserve_code_in   IN BOOLEAN DEFAULT FALSE)
   IS
      l_lines   DBMS_SQL.varchar2s;

      PROCEDURE clob_to_code_lines (
         clob_in       IN            CLOB,
         lines_inout   IN OUT NOCOPY DBMS_SQL.varchar2s)
      IS
         c_array_max         CONSTANT PLS_INTEGER DEFAULT 250;
         c_delim_not_found   CONSTANT PLS_INTEGER DEFAULT 99999999;
         c_length            CONSTANT PLS_INTEGER := LENGTH (clob_in);
         l_string                     VARCHAR2 (255);
         l_start                      PLS_INTEGER := 1;
         l_latest_delim               PLS_INTEGER;

         TYPE delims_t IS TABLE OF VARCHAR2 (1);

         l_delims                     delims_t := delims_t (' ', ';', CHR (10));

         FUNCTION latest_delim (string_in IN CLOB, start_in IN PLS_INTEGER)
            RETURN PLS_INTEGER
         IS
            l_string      VARCHAR2 (32767)
                             := SUBSTR (string_in, start_in, c_array_max);
            l_location    PLS_INTEGER := c_delim_not_found;
            l_delim_loc   PLS_INTEGER;
         BEGIN
            /* The string fits within the limits */
            IF LENGTH (l_string) < c_array_max
            THEN
               l_location := 0;
            ELSE
               /* Find the location of the last delimiter that falls before the 255 limit
                                             and use that to break up the string. */
               FOR indx IN 1 .. l_delims.COUNT
               LOOP
                  l_delim_loc :=
                     INSTR (l_string,
                            l_delims (indx),
                            -1,
                            1);

                  IF l_delim_loc > 0
                  THEN
                     l_location := LEAST (l_location, l_delim_loc);
                  END IF;
               END LOOP;

               /* If a location was found, then shift it back into the bigger string. */
               IF l_location > 0
               THEN
                  l_location := l_location + start_in - 1;
               END IF;
            END IF;

            RETURN l_location;
         END latest_delim;
      BEGIN
         IF NOT append_in
         THEN
            lines_inout.delete;
         END IF;

         WHILE (l_start <= c_length)
         LOOP
            IF preserve_code_in
            THEN
               /* Do more complex parsing. */
               l_latest_delim := latest_delim (clob_in, l_start);

               -- l_next_lf := INSTR (string_in, CHR (10), l_start);
               IF l_latest_delim = c_delim_not_found
               THEN
                  /* No delimiter found.*/
                  IF c_length - l_start + 1 > c_array_max
                  THEN
                     raise_application_error (
                        -20000,
                        'Unable to parse a CLOB without delimiters - string block too long.');
                  ELSE
                     l_string :=
                        SUBSTR (clob_in, l_start, c_array_max + l_start - 1);
                     l_start := l_start + c_array_max;
                  END IF;
               ELSIF l_latest_delim > 0
               THEN
                  l_string :=
                     SUBSTR (clob_in, l_start, l_latest_delim - l_start + 1);
                  l_start := l_latest_delim + 1;
               ELSE
                  l_string :=
                     SUBSTR (clob_in, l_start, c_array_max + l_start - 1);
                  l_start := l_start + c_array_max;
               END IF;
            ELSE
               /* Just break up the line at the c_array_max */
               l_string := SUBSTR (clob_in, l_start, c_array_max);
               l_start := l_start + c_array_max;
            END IF;

            lines_inout (lines_inout.COUNT + 1) := l_string;
         END LOOP;
      END clob_to_code_lines;
   BEGIN
      clob_to_code_lines (clob_in, l_lines);

      FOR indx IN 1 .. l_lines.COUNT
      LOOP
         DBMS_OUTPUT.put_line (l_lines (indx));
      END LOOP;
   EXCEPTION
      WHEN NO_DATA_FOUND OR VALUE_ERROR
      THEN
         /* All done! */
         NULL;
   END putline;

   /* END Substitute for DBMS_OUTPUT.PUT_LINE */
END;
 /