CREATE OR REPLACE PACKAGE loop_killer
/*
| Overview: Simple API to make it easier to insert code inside a loop
|           to check for infinite or out of control loops and kill
|           them after N iterations.
|
|           Raise the infinite_loop_detected exception.
|
|           Default "kill after": 1000 iterations
*/
IS
   c_max_iterations   CONSTANT PLS_INTEGER DEFAULT 1000;
   e_infinite_loop_detected    EXCEPTION;
   c_infinite_loop_detected    PLS_INTEGER := -20999;
   PRAGMA EXCEPTION_INIT (e_infinite_loop_detected, -20999);

   SUBTYPE loop_name_t IS VARCHAR2 (1000);

   PROCEDURE kill_after (loop_name_in        IN loop_name_t,
                         max_iterations_in   IN PLS_INTEGER);

   PROCEDURE increment_or_kill (loop_name_in   IN loop_name_t,
                                by_in          IN PLS_INTEGER DEFAULT 1);

   FUNCTION current_count (loop_name_in IN loop_name_t)
      RETURN PLS_INTEGER;
END loop_killer;
/

CREATE OR REPLACE PACKAGE BODY loop_killer
IS
   /* Need arrays for these !
   g_counter          PLS_INTEGER;
   g_max_iterations   PLS_INTEGER DEFAULT c_max_iterations;
   */

   TYPE loop_rt IS RECORD
   (
      counter          PLS_INTEGER,
      max_iterations   PLS_INTEGER
   );

   TYPE loop_t IS TABLE OF loop_rt
      INDEX BY loop_name_t;

   g_loops   loop_t;

   PROCEDURE kill_after (loop_name_in        IN loop_name_t,
                         max_iterations_in   IN PLS_INTEGER)
   IS
   BEGIN
      g_loops (loop_name_in).counter := 1;
      g_loops (loop_name_in).max_iterations :=
         NVL (GREATEST (1, max_iterations_in), c_max_iterations);
   END kill_after;

   PROCEDURE increment_or_kill (loop_name_in   IN loop_name_t,
                                by_in          IN PLS_INTEGER DEFAULT 1)
   IS
      c_message   CONSTANT VARCHAR2 (32767)
         :=    'Your loop "' || loop_name_in || '" exceeded '
            || g_loops (loop_name_in).max_iterations
            || ' iterations.' ;
   BEGIN
      g_loops (loop_name_in).counter := g_loops (loop_name_in).counter + by_in;

      IF g_loops (loop_name_in).max_iterations <= g_loops (loop_name_in).counter
      THEN
         /* Dual notification: through DBMS_OUTPUT and by raising an exception. */
         DBMS_OUTPUT.put_line ('Loop killer failure: ' || c_message);
         DBMS_OUTPUT.put_line ('Call stack below shows location of problem:');
         DBMS_OUTPUT.put_line (DBMS_UTILITY.format_call_stack);
         raise_application_error (
            c_infinite_loop_detected,
            c_message || ' Check system output for the call stack.');
      END IF;
   EXCEPTION
      WHEN NO_DATA_FOUND
      THEN
         raise_application_error (
            -20000,
            'Loop :' || loop_name_in || '" not defined.');
   END increment_or_kill;

   FUNCTION current_count (loop_name_in IN loop_name_t)
      RETURN PLS_INTEGER
   IS
   BEGIN
      RETURN g_loops (loop_name_in).counter;
   END current_count;
END loop_killer;
/

BEGIN
   loop_killer.kill_after ('outer', 3);

   LOOP
      DBMS_OUTPUT.put_line (loop_killer.current_count ('outer'));

      loop_killer.kill_after ('inner', 5);

      LOOP
         DBMS_OUTPUT.put_line (loop_killer.current_count ('inner'));
         loop_killer.increment_or_kill ('inner');
      END LOOP;

      loop_killer.increment_or_kill ('outer');
   END LOOP;
END;
/