drop table earth_parts
/

create table earth_parts (
   part_id    integer,
   part_name  varchar2(100),
   relative_size integer
)
/

begin
   insert into earth_parts values (
      100,
      'Ocean',
      1
   );

   insert into earth_parts values (
      200,
      'Mountain',
      2
   );

   insert into earth_parts values (
      300,
      'Tree',
      3
   );

   commit;
end;
/

/* Call SQL functions */

declare
   l_name1  earth_parts.part_name%type;
   l_name2  earth_parts.part_name%type;
begin
   select part_name
     into l_name1
     from earth_parts
    where part_id = 200;

   update earth_parts
      set part_name = upper(part_name)
    where part_id = 200 returning substr(part_name, 1, 5) into l_name2;

   dbms_output.put_line(l_name1
                        || '-'
                        || l_name2);
   rollback;
end;
/

/* Call built-in PL/SQL functions */

declare
   l_name1  earth_parts.part_name%type;
   l_name2  earth_parts.part_name%type;
begin
   select part_name
     into l_name1
     from earth_parts
    where part_id = 200;

   update earth_parts
      set part_name = upper(part_name)
    where part_id = 200 
    returning 'Random! ' || to_char (dbms_random.value (1, length (part_name))) 
         into l_name2;

   dbms_output.put_line(l_name1
                        || '-'
                        || l_name2);
   rollback;
end;
/

/* Call a user defined function */

create or replace function first_five (
   p_string in varchar2
) return varchar2 is
begin
   return substr(p_string, 1, 5);
end;
/

declare
   l_name1  earth_parts.part_name%type;
   l_name2  earth_parts.part_name%type;
begin
   select part_name
     into l_name1
     from earth_parts
    where part_id = 200;

   update earth_parts
      set part_name = upper(part_name)
    where part_id = 200 returning first_five(part_name) into l_name2;

   dbms_output.put_line(l_name1
                        || '-'
                        || l_name2);
   rollback;
end;
/

/* But not a nested function 

PLS-00231: function 'NESTED_FIRST_FIVE' may not be used in SQL
*/

declare
   l_name1  earth_parts.part_name%type;
   l_name2  earth_parts.part_name%type;
   function nested_first_five (
      p_string in varchar2
   ) return varchar2 is
   begin
      return substr(p_string, 1, 5);
   end;
begin
   select part_name
     into l_name1
     from earth_parts
    where part_id = 200;

   update earth_parts
      set part_name = upper(part_name)
    where part_id = 200 returning nested_first_five(part_name) into l_name2;

   dbms_output.put_line(l_name1
                        || '-'
                        || l_name2);
   rollback;
end;
/

/* Execute a query inside the returning clause! */

declare
   l_name1  earth_parts.part_name%type;
   l_name2  earth_parts.part_name%type;
begin
   update earth_parts ep
      set part_name = upper(part_name)
    where part_id = 200 returning (
      select ep2.part_name
        from earth_parts ep2
       where part_id = 200
   ),
   part_name into l_name1, l_name2;

   dbms_output.put_line(l_name1
                        || '-'
                        || l_name2);
   rollback;
end;
/

/* Watch out for mutating table errors! 

ORA-04091: table SFEUERSTEIN.EARTH_PARTS is mutating, trigger/function may not see it
*/

create or replace function first_five return varchar2 is
   l_return earth_parts.part_name%type;
begin
   select substr(part_name, 1, 5)
     into l_return
     from earth_parts
    where part_id = 200;
   return l_return;
end;
/

declare
   l_name1  earth_parts.part_name%type;
   l_name2  earth_parts.part_name%type;
begin
   update earth_parts ep
      set part_name = upper(part_name)
    where part_id = 200 returning part_name,
   (
      select first_five
        from dual
   ) into l_name1, l_name2;

   dbms_output.put_line(l_name1
                        || '-'
                        || l_name2);
   rollback;
end;
/

/* Cannot use a string indexed collection 

PLS-00657: Implementation restriction: bulk SQL with associative arrays with VARCHAR2 key is not supported.
*/

declare
   type names_t is table of earth_parts.part_name%type
      index by pls_integer;
   l_names names_t;
begin
   update earth_parts ep
      set part_name = upper(part_name)
      returning part_name bulk collect into l_names;

   dbms_output.put_line(l_names.count);
   rollback;
end;
/

declare
   type names_t is table of earth_parts.part_name%type
      index by varchar2(100);
   l_names names_t;
begin
   update earth_parts ep
      set part_name = upper(part_name)
      returning part_name bulk collect into l_names;

   dbms_output.put_line(l_names.count);
   rollback;
end;
/

/* Implicit conversions and collections */

declare
   type names_t is table of earth_parts.part_name%type
      index by pls_integer;
   l_names names_t;
begin
   update earth_parts ep
      set part_name = upper(part_name)
      returning relative_size bulk collect into l_names;

   dbms_output.put_line(l_names.count);
   rollback;
end;
/

declare
   type names_t is table of earth_parts.part_name%type
      index by pls_integer;
   l_names names_t;
   l_sizes names_t;
begin
   update earth_parts ep
      set relative_size = '1'
      returning part_name, relative_size bulk collect into l_names,l_sizes;

   dbms_output.put_line(l_names.count);
   rollback;
end;
/


declare
   type part_rt is record (nm varchar2(100), sz varchar2(100));
   
   type names_t is table of part_rt
      index by pls_integer;
   l_names names_t;
begin
   update earth_parts ep
      set relative_size = '1'
      returning part_name, relative_size bulk collect into l_names;

   dbms_output.put_line(l_names.count);
   rollback;
end;
/

/* Not with insert-select 

ORA-00933: SQL command not properly ended

workarounds:   
http://www.oracle-developer.net/display.php?id=413

*/

declare
   type names_t is table of earth_parts.part_name%type
      index by pls_integer;
   l_names names_t;
begin
   insert into earth_parts
   select * from earth_parts;
   returning part_name bulk collect into l_names;

   dbms_output.put_line(l_names.count);
   rollback;
end;
/

/* But no problem with forall - which sometimes "acts like" an insert-select */

create or replace type numlist is table of number
/

create or replace type namelist is table of varchar2 (15)
/

/* If running in LiveSQL */

create table employees as select * from hr.employees
/

create or replace procedure update_emps (col_in      in varchar2
                                       , empnos_in   in numlist)
is
   enames   namelist;
begin
   forall indx in empnos_in.first .. empnos_in.last
      execute immediate
            'UPDATE employees SET '
         || col_in
         || ' = '
         || col_in
         || ' * 1.1 WHERE employee_id = :1
         RETURNING last_name INTO :2'
         using empnos_in (indx)
         returning bulk collect into enames;

   for indx in 1 .. enames.count
   loop
      dbms_output.put_line ('10% raise to ' || enames (indx));
   end loop;
end;
/

declare
   l_ids   numlist := numlist (138, 147);
begin
   update_emps ('salary', l_ids);
end;
/

