drop table log_error_table
/

drop table err$_log_error_table
/

-- Cannot call from within SQL

create table log_error_table (error_code varchar2(10))
/

begin
   insert into log_error_table
        values (sqlcode);
end;
/

declare
   l_code   integer := sqlcode;
begin
   insert into log_error_table
        values (l_code);

   dbms_output.put_line (sql%rowcount);
end;
/

-- But OMG this works!

begin
   insert into log_error_table
        values (sys.standard.sqlcode);

   dbms_output.put_line (sql%rowcount);
end;
/

create or replace function my_sqlcode return integer
is
begin
  return sqlcode;
end;
/

begin
   insert into log_error_table
        values (my_sqlcode);

   dbms_output.put_line (sql%rowcount);
end;
/

create or replace function my_sqlcode return integer
is
   pragma autonomous_transaction;
begin
  return sqlcode;
end;
/

begin
   insert into log_error_table
        values (my_sqlcode);

   dbms_output.put_line (sql%rowcount);
end;
/

-- sqlcode 0 unless called from within exception handler. That's runtime within!

create or replace procedure show_sqlcode is
begin
   dbms_output.put_line(sqlcode);
end;
/

begin
   show_sqlcode;
   raise value_error;
exception
   when others then
      show_sqlcode;
end;
/

-- an error with two error codes?

declare
   l_dummy dual.dummy%type;
begin
   select dummy into l_dummy from dual
    where 1 = 2;
exception
   when others then
      dbms_output.put_line('sqlcode = ' || sqlcode);
end;
/

declare
   l_list dbms_sql.varchar2_table;
begin
   dbms_output.put_line('l_list(1) = ' || l_list(1));
exception
   when others then
      dbms_output.put_line('sqlcode = ' || sqlcode);
end;
/

declare
   l_list dbms_sql.varchar2_table;
begin
   dbms_output.put_line('l_list(1) = ' || l_list(1));
end;
/

-- but at least sqlcode and sqlerrm are consistent....?

declare
   l_list dbms_sql.varchar2_table;
begin
   dbms_output.put_line('l_list(1) = ' || l_list(1));
exception 
   when others then
      dbms_output.put_line('sqlcode = ' || sqlcode);
      dbms_output.put_line('sqlerrm = ' || sqlerrm);
end;
/

-- and you better remember to use the right one.

declare
   e exception;
   pragma exception_init (e, -1403); -- -140444444444);
begin
   dbms_output.put_line ('hurray!');
end;
/

-- ok but we all know that error codes are (generally) negative, right?

-- except for 1 and 100 I mean.

declare
   e exception;
begin
   raise e;
exception 
   when others then
      dbms_output.put_line('sqlcode = ' || sqlcode);
      dbms_output.put_line('sqlerrm = ' || sqlerrm);
end;
/

begin
   dbms_errlog.create_error_log (dml_table_name => 'LOG_ERROR_TABLE');
end;
/

insert into log_error_table values ('This is not an error code')
/

insert into log_error_table values ('This is not an error code')
log errors reject limit unlimited
/

select * from err$_log_error_table
/

-- and as for sql%bulk_exceptions...sigh...

declare
   l_codes dbms_sql.varchar2_table;
begin
   l_codes (1) := 'a really long error code';
   forall indx in 1 .. l_codes.count
      save exceptions
      insert into log_error_table values (l_codes(indx));
exception
   when others
   then
      for indx in 1 .. sql%bulk_exceptions.count
      loop
         dbms_output.put_line (sql%bulk_exceptions (indx).error_code);
      end loop;
end;      
/   

-- but what about 1 and -1?

create or replace function fail_with_dvoi return varchar2
is
begin
   raise dup_val_on_index;
end;
/

create or replace trigger fail_on_insert_tr
before insert on log_error_table
for each row
begin
   :new.error_code := fail_with_dvoi();
end;
/

insert into log_error_table values (-1)
log errors reject limit unlimited
/

select * from err$_log_error_table
/

create or replace function fail_with_ude return varchar2
is
   e exception;
begin
   raise e;
end;
/

create or replace trigger fail_on_insert_tr
before insert on log_error_table
for each row
begin
   :new.error_code := fail_with_ude();
end;
/

insert into log_error_table values (-1)
log errors reject limit unlimited
/

select * from err$_log_error_table
/

